/*!
 *  @header Daubechies.h
 *  
 *  Created by Pierre Derian on 20/10/09.
 *  @copyright 2009 INRIA. All rights reserved.
 */

#include <iostream>
#include "CImg.h"

#ifndef DEF_DAUBECHIES
#define DEF_DAUBECHIES

/*!
 * @class       Daubechies
 * @abstract    This class implement Daubechies wavelet transform and various wavelet-based operations.
 */
class Daubechies {

  /*===============================*/
  /*            METHODS            */
  /*===============================*/
  
public:
  /*!
   * @abstract  default constructor
   */
  Daubechies();
  
  /*!
   * @abstract  normal contructor
   * @param     P number of vanishing moments (VM)
   * @param     type 0 for Daubechies, 1 for Coiflets (even VM only!)
   */
  Daubechies(int P, int type=0);
  
  /*!
   * @abstract  Gets the number of vanishing moments of instance.
   * @result    The number of vanishing moments.
   */
  int get_nMoments() const;
  
  /*!
   * @abstract  Gets the wavelet bases type (0: Daubechies, 1:Coiflet).
   * @result    The type.
   */
  int get_type() const;
  
  /*!
   * @abstract  Gets the filter h.
   * @result    The filter h.
   */
  cimg_library::CImg<double> get_h() const;
  
  /*!
   * @abstract  Gets the filter g.
   * @result    The filter g.
   */
  cimg_library::CImg<double> get_g() const;
  
  /*!
   * @abstract  Displays wavelet type and number of vanishing moments.
   * @param     printFilter to enable print of h & g filter values.
   */
  void info(bool printFilters=false);

  /*!
   * @abstract  Implements complete wavelet transform for 1D (row) image.
   * @param     image Input image.
   * @param     level Number of levels of decomposition.
   * @result    Coefficients from decomposition.
   */
  cimg_library::CImgList<double> image2coeff1D(const cimg_library::CImg<double>& image, int level);
  
  /*!
   * @abstract  Implements complete wavelet transform for 2D images.
   * @param     image Input image.
   * @param     level Number of levels of decomposition.
   * @result    Coefficients from decomposition.
   */  
  cimg_library::CImgList<double> image2coeff2D(const cimg_library::CImg<double>& image, int level);

  /*!
   * @abstract  Rebuilds a 1D image from its coefficents.
   * @param     coeff List of coefficients.
   * @param     level Number of levels of reconstruction.
   * @result    the associated image.
   */
  cimg_library::CImg<double> coeff2image1D(const cimg_library::CImgList<double>& coeff, int level);
  
  /*!
   * @abstract  Rebuilds a 2D image from its coefficents.
   * @param     coeff List of coefficients.
   * @param     level Number of levels of reconstruction.
   * @result    the associated image.
   */
  cimg_library::CImg<double> coeff2image2D(const cimg_library::CImgList<double>& coeff, int level);
  
  /*!
   * @abstract  Rebuilds a 2D image from its coefficents.
   * @discussion "coarse" version is interesting when several details scales are empty (=0).
   *            It reduces the computational cost.
   * @param     coeff List of coefficients.
   * @param     level Number of levels of reconstruction.
   * @param     detailLimit is the first scale having null details.
   * @result    the associated image.
   */  
  cimg_library::CImg<double> coarseCoeff2image2D(const cimg_library::CImgList<double>& coeff, int level, int detailLimit);
  
  /*!
   * @abstract  Apply successive decompositions (with filterbank) on 1D signal.
   * @discussion  decompose1DI() returns a 1D-CImg (row) instead of CImgList.
   * @param     signal Input signal.
   * @param     level Number of successive decompositions (>=0). Must not be bigger than the signal resolution, e.g. with 256=2^8pixels : 0<=level<=8.
   * @result    Coefficients from the decomposition as a list of images - see @link compose1D compose1D@/link for explanations.
   */
  cimg_library::CImgList<double> decompose1D(const cimg_library::CImg<double>& signal, int level);
  
  /*!
   * @abstract  Apply successive decompositions (with filterbank) on 1D signal. CImg-based version (instead of CImgList).
   * @param     signal Input signal.
   * @param     level Number of successive decompositions (>=0). Must not be bigger than the signal resolution, e.g. with 256=2^8pixels : 0<=level<=8.
   * @result    Coefficients from the decomposition stored within a single image.
   */  
	cimg_library::CImg<double> decompose1DI(const cimg_library::CImg<double>& signal, int level);
	cimg_library::CImg<double> decompose1DI(const cimg_library::CImg<double>& signal, int level, double zeta);
 	
  /*!
   * @abstract  Rebuilds a 1D signal.
   * @param     data This is a CImgList containing coefficients for each level:
   *            data(0) coarse approximation;
   *            data(k), k>0 are the (k-1)th scale details coefficients.
   *            They must be stored in as rows.
   * @param     Roffset Contains the (eventual) index offset of the result. 
   * @param     level The number of iterations to perform.
   * @param     border 0 if signal is infinite, 1 (default) if finite (uses periodization).
   * @result    The resulting approximation. 
   */
  cimg_library::CImg<double> compose1D(const cimg_library::CImgList<double>& data, int& Roffset, int level, int border=1);
  
  /*!
   * @abstract  Rebuilds a 1D signal. CImg-based version (instead of CImgList).
   * @param     data 1D-CImg (row) containing coefficients.
   * @param     Clevel the coarse reolution (>=0).
   * @result    The resulting approximation. 
   */
	cimg_library::CImg<double> compose1DI(const cimg_library::CImg<double>& data, int Clevel);
	cimg_library::CImg<double> compose1DI(const cimg_library::CImg<double>& data, int Clevel, double zeta);
  
  /*!
   * @abstract  Apply successive decompositions (with filterbank) on a 2D signal.
   * @param     signal Input 2D signal.
   * @param     level Number of successive decompositions (>=0). Must not be bigger than the signal resolution, e.g. with 256=2^8pixels : 0<=level<=8.
   * @result    Coefficients from the decomposition as a list of images - see @link compose2D compose2D@/link for explanations.
   */
  cimg_library::CImgList<double> decompose2D(const cimg_library::CImg<double>& signal, int &level);
  
  /*!
   * @abstract  Rebuilds a 2D signal, using periodization.
   * @param     data This is a CImgList containing coefficients for each level:
   *            data(0) = a, coarse level values;
   *            data(k)(slices 0,1,2), k>0 are the (k-1)th details coefficients. 3 slices per scale : horizontal (0), vertical(1) and diagonal(2) coefficients.
   * @param     level The number of iterations to perform.
   * @result    The resulting approximation.
   */
  cimg_library::CImg<double> compose2D(const cimg_library::CImgList<double>& data, int level);
  
  /*!
   * @abstract  Rebuilds a 2D signal with EMPTY DETAILS, using periodization.
   * @param     coarse the coarse approximation
   * @param     level number of iterations.
   * @result    The resulting approximation.
   */
  cimg_library::CImg<double> coarseCompose2D(const cimg_library::CImg<double>& coarse, int level);

	cimg_library::CImg<double> anisDecompose2D(const cimg_library::CImg<double>& signal,
											   int &level);
	cimg_library::CImg<double> anisDecompose2D(const cimg_library::CImg<double>& signal,
											   int &level, bool normalize);
	
	cimg_library::CImg<double> anisDecompose2DFracNormalize(const cimg_library::CImg<double>& signal,  
											   int &level, double &zeta);
  
	cimg_library::CImg<double> anisCompose2D(const cimg_library::CImg<double>& data,
											 int &level);
	cimg_library::CImg<double> anisCompose2D(const cimg_library::CImg<double>& data,
											 int &level, bool normalize);
	
	cimg_library::CImg<double> anisCompose2DFracNormalize(const cimg_library::CImg<double>& data, 
											 int &level, double &zeta);
	
/*!
   * @abstract  Computes values of the wavelet/scale function associated to filters g/h.
   * @param     Roffset (out) Contains the (eventual) index offset of the result.
   * @param     level The number of iterations to perform, also the required precision:
   *            2^level samples over the unit segment.
   * @param     func 0 for the scale function, 1 for the wavelet.
   * @result    Values of the chosen function over its support.
   */
  cimg_library::CImg<double> wFunc(int& Roffset, int level, int func);
  
  /*!
   * @abstract  Computes the values of the (eventually scaled and translated) scaling function Phi at a given resolution. 
   * @discussion Designed for @link projection1D projection1D@/link / @link projection2D 2D@/link.
   * @param     totLvl is the resolution level.
   * @param     resLvl
   * @param     n is the translation.
   * @result    Function value.
   */
  cimg_library::CImg<double> phi(int totLvl, int resLvl, int n=0);
  
  /*!
   * @abstract  Computes the values of the (eventually scaled and translated) wavelet function Psi at a given resolution. 
   * @param     totLvl is the resolution level.
   * @param     resLvl
   * @param     n is the translation.
   * @result    Function value.
   */
  cimg_library::CImg<double> psi(int totLvl, int resLvl, int n=0);
  
  /*!
   * @abstract  Compute values of (1D) periodized scaling and wavelet functions at chosen resolution, along with its translations and scalings.
   * @param     Phi (out) Scaling function values.
   * @param     Psi (out) Wavelet function values.
   * @param     nLevel the resolution (2^nLevel samples on [0,1[).
   */
  void listWavelets(cimg_library::CImgList<double> &Phi, cimg_library::CImgList<double> &Psi, int nLevel);
  
  /*!
   * @abstract  Computes the projection of a 1D picture onto higher resolution spaces.
   * @param     In Input picture.
   * @param     level The number of successive projection(s).
   * @result    The projected image.
   */
  cimg_library::CImg<double> projection1D(const cimg_library::CImg<double>& In,
                                          int level);
  /*!
   * @abstract  Computes the projection of a 1D picture, from its coefficients, into higher resolution spaces.
   * @param     cIn Input coefficient pyramid.
   * @param     level The number of successive projection(s).
   * @result    The projected image.
   */  
  cimg_library::CImg<double> projection1D(const cimg_library::CImgList<double>& cIn,
                                          int level);
  
  /*!
   * @abstract  Computes the projection of a 2D picture into higher resolution space.
   * @param     In Input picture.
   * @param     level The number of successive projection(s).
   * @result    The projected image.
   */
  cimg_library::CImg<double> projection2D(const cimg_library::CImg<double>& In,
                                          int level);
  /*!
   * @abstract  Computes the projection of a 2D picture, from its coefficients, into higher resolution space.
   * @param     cIn Input coefficient pyramid.
   * @param     level The number of successive projection(s).
   * @result    The projected image.
   */
  cimg_library::CImg<double> projection2D(const cimg_library::CImgList<double>& cIn,
                                          int level);
  
  /*!
   * @abstract  Interpolation 1D with wavelets. Periodic boundary conditions.
   * @param     In is the input 1D signal - MUST BE A ROW !
   * @result    The interpolated image (twice bigger).
   */
  cimg_library::CImg<double> interp1D(const cimg_library::CImg<double>& In);
  
  /*!
   * @abstract  Interpolation 2D with wavelets. Periodic boundary conditions.
   * @param     In is the input 2D signal.
   * @result    The interpolated image (twice bigger in each direction).
   */
  cimg_library::CImg<double> interp2D(const cimg_library::CImg<double>& In);
  
  /*!
   * @abstract  Computes diagonaly-interpolated signal.
   * @discussion This is like a "subset" of @link interp2D interp2D@/link. 
   * @param     In is the input 2D signal.
   * @result    Interpolated image.
   */
  cimg_library::CImg<double> diagInterp2D(const cimg_library::CImg<double>& In);
  
  /*!
   * @abstract  Transposed application of @link diagInterp2D diagInterp2D@/link.
   * @param     In is the input 2D signal.
   * @result    "TranspInterpolated" image.
   */
  cimg_library::CImg<double> diagTranspInterp2D(const cimg_library::CImg<double>& In);
  
  /*!
   * @abstract  [DEV STUFF!] Build the "classical" representation of a wavelet decomposition of a given image.
   * @discussion I.e. returns the approximation surrounded by the different details scales.
   *            All 3 subsets of coefficients are normalized at a given scale to help visualization.
   *            /!\ as a consequence it does not reflect their true amplitude.
   * @param     In is the input image
   * @param     level is the number of decomposition level (or detail scales).
   * @result    Approximation and details appened.
   */
  cimg_library::CImg<double> niceView(const cimg_library::CImg<double>& In, int level);
  
  /*!
   * @abstract  [WIP!] Computes isotropic wavelet-spectrum from given (2D) coefficients.
   * @param     coeffs The coefficients
   * @result    an Nx2 images, where N is the number of scales in the coefficients pyramid,
   *            first row is scales and second row is corresponding spectrum values.
   */
  cimg_library::CImg<double> iso2DWavSpectrum(const cimg_library::CImgList<double> &coeffs); 
  
protected:
  
  /*!
   * @abstract  Sets up filters (and their offsets) for Daubechies wavelets.
   */
  void setDaubechiesFilters();
  
  /*!
   * @abstract  Sets up filters (and their offsets) for Coiflet wavelets.
   */
  void setCoifletFilters();
  
  /*!
   * @abstract  Performs one step of the decomposition filter bank.
   * @discussion Returns a CImgList of two images, where
   *
   *            result(0) := (>2)[a*hb]
   *            result(1) := (>2)[a*gb]
   *
   *            /!\ Does not deal with offsets, i.e. only for signal a[n], n>=0.
   * @param     a Signal to be processed - must be a row!
   * @result    A CImgList of 1-row images, image 0 is coarse and 1 is details.
   */
  cimg_library::CImgList<double> stepDown(const cimg_library::CImg<double>& a);
  
  /*!
   * @abstract  Performs one step of the decomposition filter bank. CImg-based version (instead of CImgList).
   * @discussion Coarse approximation and details are stored consecutively. If input signal has N elements, then:
   *
   *            result(0, ..., N/2-1) := (>2)[a*hb]
   *            result(N/2, ..., N-1) := (>2)[a*gb]
   *
   *            /!\ Does not deal with offsets, i.e. only for signal a[n], n>=0.
   * @param     a Signal to be treated - must be a row!
   * @result    A 1-row image, first half is coarse and second half is details.
   */
  cimg_library::CImg<double> stepDownI(const cimg_library::CImg<double>& a);
  
  /*!
   * @abstract  Performs one step of the reconstruction filter bank.
   * @discussion i.e.
   *
   *            res := (<2)[a*h] + (<2)[b*g]
   * @param     a Signal to be expanded and convolved with h  - must be a row!
   * @param     Aoffset is the eventual index offset for a. 
   * @param     b Signal to be expanded and convolved with g - also a row!
   * @param     Boffset is the eventual index offset for b.
   * @param     Roffset [out] Contains the (eventual) index offset of the result. 
   * @param     border 0 if signal is infinite, 1 if finite (uses periodization).
   * @result    The resulting coefficients.
   */
  cimg_library::CImg<double> stepUp(const cimg_library::CImg<double>& a, const int& Aoffset,
              const cimg_library::CImg<double>& b, const int& Boffset,
              int& Roffset, int border);
  
  /*!
   * @abstract  Performs one step of the reconstruction filter bank. Simpler version for simpler use!
   * @discussion 
   *
   *            result := (<2)[a*h] + (<2)[b*g]
   *            This version is intented for real signals (classical use), so we neglect offsets (always 0).
   * @param     a Signal to be expanded and convolved with h  - must be a row!
   * @param     b Signal to be expanded and convolved with g - also a row!
    * @param     border 0 if signal is infinite, 1 if finite (uses periodization).
   * @result    The resulting coefficients.
   */
  cimg_library::CImg<double> stepUpI(const cimg_library::CImg<double>& a,
                                    const cimg_library::CImg<double>& b);
  /*!
   * @abstract  Performs one step of the reconstruction filter bank, with NULL DETAIL COEFFICIENTS,
   * @discussion i.e.
   *
   *            res := (<2)[a*h]
   * @param     a Signal to be expanded and convolved with h  - must be a row!
   * @param     Aoffset is the eventual index offset for a. 
   * @param     Roffset [out] Contains the (eventual) index offset of the result. 
   * @param     border 0 if signal is infinite, 1 if finite (uses periodization).
   * @result    The resulting coefficients.   
   */
  cimg_library::CImg<double> coarseStepUp(const cimg_library::CImg<double>& a, const int& Aoffset,
                                    int& Roffset, int border);
  
   
  /*===============================*/
  /*     STATIC METHODS            */
  /*===============================*/
  
public:
  
  /*!
   * @abstract  Returns a coefficients pyramid initialized to 0 for the given image (2D) dimensions and number of decomposition levels
   * @param     width image horizontal dimension
   * @param     height image vertical dimension
   * @param     level number of decomposition
   * @result    A list of images (2D)
   **/
  static cimg_library::CImgList<double> allocCoeffs(int width, int height, int &level);
  
  /*!
   * @abstract  Regroups coefficients from a 2D decomposition (CImgList) into a single image, with the 'classical' dyadic structure.
   * @discussion Horizontal coeff are put upright, vertical ones downleft, diagonal ones downright. In WaveLab, vertical and horizontal are swaped. 
   * @param     pyramid A CImgList containing coefficients, usually gotten from decompose2D() or image2coeff2D()
   * @result    An image containing coeff values
   **/
  static cimg_library::CImg<double> regroupCoeffPyramid(const cimg_library::CImgList<double> &pyramid);
  
  /*!
   * @abstract  Computes and returns wavelet leaders of 1D signal coefficients, according to Jaffard definition.
   * @param     wavCoeffs Input coefficients obtained from signal decomposition.
   * @param     vanishMap [out] indicates wether a coefficent has to be taken into account in the structure function computation, are not (if it vanishes). 
   * @param     startLevel Scales from startLevel to lastest level (i.e. smallest scales) will be explored.
   * @result    Returns wavelet leaders at different scales.
   */
  static cimg_library::CImgList<double> get_Leaders1D(const cimg_library::CImgList<double>& wavCoeffs, cimg_library::CImgList<double>& vanishMap, int startLevel);
  
  /*!
   * @abstract  Computes and returns wavelet leaders of 2D signal coefficients, according to Jaffard definition.
   * @param     wavCoeffs Input coefficients obtained from signal decomposition.
   * @param     vanishMap [out] indicates wether a coefficent has to be taken into account in the structure function computation, are not (if it vanishes). 
   * @param     startLevel Scales from startLevel to lastest level (i.e. smallest scales) will be explored.
   * @result    Returns wavelet leaders at different scales.
   */
  static cimg_library::CImgList<double> get_Leaders2D(const cimg_library::CImgList<double>& wavCoeffs, cimg_library::CImgList<double>& vanishMap, int startLevel);
  
  /*!
   * @abstract  Computes structure functions from 1D wavelet leaders.
   * @param     Coeffs Wavelet leaders.
   * @param     vanishMap The associated vanishing coefficients map, to ensure stability for negative orders.
   * @param     order is the maximum order to be evaluated : orders between [-order, order] will be conmputed.
   * @param     nVal is the number of samples desired within [-order, order].
   * @result    An array containing structure functions values for [order][scale].
   */
  static double** get_StructFunc1D(const cimg_library::CImgList<double>& Coeffs, const cimg_library::CImgList<double>& vanishMap, int order, int nVal);
  
  /*!
   * @abstract  Computes structure functions from 2D wavelet leaders.
   * @param     Coeffs Wavelet leaders.
   * @param     vanishMap The associated vanishing coefficients map, to ensure stability for negative orders.
   * @param     order is the maximum order to be evaluated : orders between [-order, order] will be computed.
   * @param     nVal is the number of samples desired within [-order, order].
   * @param     type is the type of coefficients to be analysed (0 : horizontal, 1 : vertical).
   * @result    An array containing structure functions values for [order][scale].
   */
  static double** get_StructFunc2D(const cimg_library::CImgList<double>& Coeffs, const cimg_library::CImgList<double>& vanishMap, int order, int nVal, int type); 
  
  /*!
   * @abstract  Subfunction for @link get_Leaders1D get_Leaders1D@/link.
   * @param     Coeffs Abs() of input coefficients.
   * @param     vanishMap [out].
   * @param     level Scale to be treated
   * @param     q Coordinate for Leader search.
   * @result    Value of the dominating coeff at scale level, index q.
   */
  static double findLeaders1D(const cimg_library::CImgList<double>& Coeffs, cimg_library::CImgList<double>& vanishMap, int level, int q);
  
  /*!
   * @abstract  Subfunction for @link get_Leaders2D get_Leaders2D@/link.
   * @param     Coeffs Abs() of input coefficients.
   * @param     vanishMap [out].
   * @param     level Scale to be treated
   * @param     p
   * @param     q
   * @param     z Coordinates for Leader search.
   * @result    Value of the dominating coeff at scale level, index p,q,z.
   */
  static double findLeaders2D(const cimg_library::CImgList<double>& Coeffs, cimg_library::CImgList<double>& vanishMap, int level, int p, int q, int z);
  
protected:  
  
  /*!
   * @abstract  Recursive function returning supp{ coeff } of the dyadic 1D tree beginning at given coordinates.
   * @param     Coeffs Abs() of input coefficients.
   * @param     level Top scale for dyadic tree exploration.
   * @param     q Coordinate of the beginning of the dyadic tree.
   * @result    l_lambda = sup { c_lambda' }, with lambda' such as supp(Psi_lambda') included in 3*supp(Psi_lambda). 
   */
  static double lookBelow1D(const cimg_library::CImgList<double>& Coeffs, int level, int q);

  /*!
   * @abstract  Recursive function returning supp{ coeff } of the dyadic 2D tree beginning at given coordinates.
   * @param     Coeffs Abs() of input coefficients.
   * @param     level Top scale for dyadic tree exploration.
   * @param     p
   * @param     q   
   * @param     z Coordinates of the beginning of the dyadic tree.
   * @result    l_lambda = sup { c_lambda' }, with lambda' such as supp(Psi_lambda') included in 3*supp(Psi_lambda). 
   */  
  static double lookBelow2D(const cimg_library::CImgList<double>& Coeffs, int level, int p, int q, int z);

  /*
   *  period
   */
  static inline int period (int ind, int max);
  
  /*!
   * @abstract  Inserts 0 in the input signal.
   * @discussion Notice that the associated offset will have to be doubled.
   * @param     A The input signal - not modidied.
   * @result    Expanded signal.
   */
  static cimg_library::CImg<double> expand(const cimg_library::CImg<double>& A);
  
  /*!
   * @abstract  Keeps only one value out of 2 in input signal.
   * @discussion /!\ only with signal a[n], n>=0 /!\
   * @param     A input signal.
   * @result    Decimated signal.
   */
  static cimg_library::CImg<double> decimate(const cimg_library::CImg<double>& A);
  
  /*!
   * @abstract  Performs 1D convolution.
   * @param     signal Signal to be convolved - must be a row!
   * @param     Soffset is the eventual index offset for signal. 
   * @param     filtre The filter - also a row!
   * @param     Foffset is the eventual index offset for F.
   * @param     Roffset [out] Contains the (eventual) index offset of the result. 
   * @param     border 0 if signal is infinite, 1 if finite (uses periodization).
   * @result    The resulting coefficients.
   */
  static cimg_library::CImg<double> convolve(const cimg_library::CImg<double>& signal, const int& Soffset, 
                                             const cimg_library::CImg<double>& filtre, const int& Foffset, 
                                             int& Roffset, int border);
  
  /*!
   * @abstract  Adds two coefficients rows, taking their respective offsets into account.
   * @param     a First row.
   * @param     Aoffset Associated offset.
   * @param     b Second row.
   * @param     Boffset Associated offset.
   * @param     Roffset [out] Result offset.
   * @result    The sum!
   */
  static cimg_library::CImg<double> add(const cimg_library::CImg<double>& a, const int& Aoffset,
                                        const cimg_library::CImg<double>& b, const int& Boffset,
                                        int& Roffset);
  
  /*===============================*/
  /*            MEMBERS            */
  /*===============================*/
  
private:
  /*! @var nMoments Number of vanishing moments. */
  int nMoments;
  /*! @var wavType wavelet type. */
  int wavType;  
  /*! @var h Filter h (associated to scale function). */
  cimg_library::CImg<double> h;
  /*! @var hb Filter \bar(h) (conjugate mirror filter). */
  cimg_library::CImg<double> hb;
  /*! @var g Filter g (associated to wavelet function). */  
  cimg_library::CImg<double> g;
  /*! @var gb Filter \bar(g) (conjugate mirror filter). */  
  cimg_library::CImg<double> gb;
  /*! @var hi Interpolation filter. */  
  cimg_library::CImg<double> hi;
  /*! @var hib Conjugate Mirror Interpolation filter (for transposed application). */  
  cimg_library::CImg<double> hib;
  /* offsets of filters */
  int hOff, gOff, hbOff, gbOff, hiOff, hibOff;
};

#endif