/*
 *  OpFlowTools_io.h
 *  
 *
 *  Copyright 2014 INRIA. All rights reserved.
 *
 */

#ifndef DEF_OFTOOLS_IO
#define DEF_OFTOOLS_IO

#include "CImg.h"
#include <fstream>
#include <iostream>
#include <iomanip>

/*===========GENERIC FUNCTIONS===============*/

/**
 *  loadMotion
 *
 * Automatically loads .motion (text), .inr (binary), or nc velocity files, according to extension.
 */
cimg_library::CImg<double> loadMotion(const char *const filename);

/**
 *  loadIm
 *
 * Automatically loads image files (including netcdf files), according to extension.
 */
cimg_library::CImg<double> loadIm(const char *const filename);

/**
 *  saveMotion
 *
 * Automatically saves .motion (text) or .inr (binary) velocity files, according to extension.
 */
int saveMotion(const char *const filename, const cimg_library::CImg<double> &UV);


/*===========DEDICATED FUNCTIONS===============*/

/**
 *  write_NCDF
 *
 * Writes a standard NCDF file.
 * Rem: requires nclib.
 */
#ifdef USE_NCLIB
int write_NCDF(const char *const filename,
              const cimg_library::CImg<double> &UV);
#endif

/**
 *  read_NCDF
 *
 * Reads a NCDF file of variables u, v (x and y are omitted), of dimensions x and y.
 * Rem: requires nclib.
 */
#ifdef USE_NCLIB
cimg_library::CImg<double> read_NCDF(const char *const filename);
cimg_library::CImg<double> read_NCDF_Im(const char *const filename);
#endif
                         
/**
 *  read_aUVFile
 *
 * Build a velocity field image from the given file. 
 */
cimg_library::CImg<double> read_aUVFile(char * filename,
                                        int dx,
                                        int dy,
                                        int Dx,
                                        int Dy,
                                        int ofx,
                                        int ofy);
/**
 *  write_aUVFile
 *
 */
template <typename Tfloat> void write_aUVFile(char * filename,
                                              const cimg_library::CImg<Tfloat> &UV){
  
  std::fstream file(filename, std::ios::out | std::ios::trunc); 
  //-initiate aux with UV
	cimg_library::CImg<double> aux(UV);
  //-mirror aux
  aux.mirror('y');
  
  if (file){
    //--write file column by column
    for (int j=0;j<aux.width();j++){
      for (int i=0;i<aux.height();i++){
        file << j << " " << i << " " << std::setprecision(9) << aux(j,i,0) << " " << std::setprecision(9) << -aux(j,i,1) << std::endl;
      }
      file << std::endl;
    }
    file.close();
    std::cout << "file " << filename << " successfully written." << std::endl;
  }
  else {
    std::cout << "\t\t/!\\ impossible to open file for writting." << std::endl;
  }
};


/**
 *  read_MotionFile
 */

/*- Cimg version */
cimg_library::CImg<double> read_MotionFile(const char *const filename,
                                           int &err);
/*- double[] version -*/
void read_MotionFile(const char *const filename,                         
                     int &dx,
                     int &dy,
                     double* &field_u,
                     double* &field_v,
                     int &err);

/*- header processing -*/
void read_MotionFileHeader(const char *const filename,                         
                           int &dx,
                           int &dy,
                           int &err);
/*- data reading -*/
void read_MotionFileData(const char *const filename,                         
                         int dx,
                         int dy,
                         double* field_u,
                         double* field_v,
                         int &err);

/**
 * write_MotionFile
 */
template<typename Tfloat> void write_MotionFile(const char *const filename,
                                                int dx,
                                                int dy,
                                                Tfloat *U,
                                                Tfloat *V,
                                                int &err){
  
  std::fstream file(filename, std::ios::out | std::ios::trunc);
  if (file){
    file << dx << " " << dy << std::endl;
    file << "#i\tj\tU\tV" << std::endl;
    for (int p=0; p<dy; p++) {
      for (int q=0; q<dx; q++) {
        file << p << "\t" << q << "\t" << std::setiosflags(std::ios::fixed) << std::setprecision(6) << U[p*dx+q] << "\t" << std::setiosflags(std::ios::fixed) << std::setprecision(6) << V[p*dx+q] << std::endl;
      }
    }
    file.close();
  }
  else {
    std::cout << "Impossible to open file " << filename << ". Exit !" << std::endl;
    err = 2;
  }
  
};
template<typename Tfloat> void write_Spectrum(const char *const filename,
											    int dx,
                                                Tfloat *U,
                                                int &err){
	
	std::fstream file(filename, std::ios::out | std::ios::trunc);
	if (file){
		file << "#kMax="<<dx<< std::endl;
		file << "#k^2\tE(k)" << std::endl;
		for (int q=0; q<dx; q++) {
			file << U[q]<< "\t" << std::setiosflags(std::ios::scientific) << std::setprecision(15) << U[q+dx]<< std::endl;
		}
		file.close();
	}
	else {
		std::cout << "Impossible to open file " << filename << ". Exit !" << std::endl;
		err = 2;
	}
	
};

/**
 *  read_pgm
 */
cimg_library::CImg<short> read_pgm(char *nom,
                                   int dx,
                                   int dy,
                                   int ofx,
                                   int ofy);

/**
 *  read-lidar
 */
void read_lidar(char* filename,
                double* output,
                int width,
                int height,
                int &err);

#endif