/*
 *  OpFlowTools_io.cpp
 *  
 *
 *  Copyright 2014 INRIA. All rights reserved.
 *
 */

#include "OpFlowTools_io.h"
#ifdef USE_NCLIB  
  #include "netcdfcpp.h"
#endif  

using namespace cimg_library;
using namespace std;

/**
 *  Generic function to load .motion (text), .inr or .nc (binaries)
 */
CImg<double> loadMotion(const char *const filename){
  
  bool loaded;
  CImg<double> UV;
  int err = 0;
  const char *const ext = cimg::split_filename(filename);
  
  //-load according to extension
  if (!cimg::strcasecmp(ext,"motion")){
    UV.assign(read_MotionFile(filename,err));
    loaded = (err)?false:true;
  }
  else if (!cimg::strcasecmp(ext,"inr")){
    UV.load_inr(filename);
    loaded = true; //-if failure, it crashes...
  }
  #ifdef USE_NCLIB
  else if (!cimg::strcasecmp(ext,"nc")){
    UV = read_NCDF(filename);
    loaded = true;
  }
#endif  
  else{
    cout << "/!\\ Unknown extension in file "<< filename <<". Use .motion or .inr!" << endl;
    loaded = false;
  }
  //-warn if we failed to load
  if (!loaded){cout << "/!\\ Impossible to load " << filename <<", empty field returned..." << endl;}
  
  return UV;
};


CImg<double> loadIm(const char *const filename){
    

    CImg<double> Im;
    const char *const ext = cimg::split_filename(filename);
    
    //-load according to extension
    if (cimg::strcasecmp(ext,"nc")){
        Im.assign(filename);
    }
#ifdef USE_NCLIB
    else {
        Im.assign(read_NCDF_Im(filename)).mirror('y'); ;
    }
#endif
    
    return Im;
};


/**
 *  Generic function to save .motion (text), .inr or .nc (binaries)
 * written on 14/01/2011.
 * updated 24/01/2011 (.nc)
 */
int saveMotion(const char *const filename, const CImg<double> &UV){
  
  int err = 0;
  const char *const ext = cimg::split_filename(filename);
  
  //-save according to extension
  if (!cimg::strcasecmp(ext,"motion")){
    write_MotionFile(filename,UV.width(),UV.height(),UV.get_shared_plane(0).data(),UV.get_shared_plane(1).data(),err);
  }
  else if (!cimg::strcasecmp(ext,"inr")){
    //-first (crudly) cast to a float version
    CImg<float>(UV).save_inr(filename);
  }
#ifdef USE_NCLIB
  else if (!cimg::strcasecmp(ext,"nc")){
    write_NCDF(filename,UV);
  }
#endif  
  else{
    cout << "/!\\ Unknown extension in file "<< filename <<". Use .motion or .inr!" << endl;
    err = 1;
  }
  //-warn if we failed to save
  if (err){cout << "/!\\ Impossible to save " << filename <<" ... Sorry about that ^_^." << endl;}
  return err;
}

/*========READ or WRITE IMAGES / FIELDS=============*/

/**
 *  Write .nc files.
 *
 * written on 24/01/2011
 */
#ifdef USE_NCLIB
int write_NCDF(const char *const filename,
              const CImg<double> &UV){
  
  int err = 0;
  int nx, ny;
  float *x, *y;
  CImg<float> tempUV(UV, false);
  
  //-re orient
  tempUV.get_shared_plane(1) *= -1.;
  
  //-open file
  NcFile dataFile(filename, NcFile::Replace);
  if (dataFile.is_valid())
  {
    //-grid data
    nx = UV.width();
    ny = UV.height();
    x = new float[nx];
    y = new float[ny];
    for (int k=0; k<nx; k++) x[k] = (float)k;
    for (int k=0; k<ny; k++) y[k] = (float)k;
    //-create dimensions
    NcDim *xDim, *yDim;
    xDim = dataFile.add_dim("x", nx);
    yDim = dataFile.add_dim("y", ny);
    //-create grid points (x, y)
    NcVar *dataX, *dataY;
    dataX = dataFile.add_var("x", ncFloat, xDim);
    dataY = dataFile.add_var("y", ncFloat, yDim);
    //-create variable u and v
    NcVar *dataU, *dataV;
    dataU = dataFile.add_var("u", ncFloat, xDim, yDim);
    dataV = dataFile.add_var("v", ncFloat, xDim, yDim);
    //-units
    dataX->add_att("units","pixels");
    dataY->add_att("units","pixels");
    dataU->add_att("units","pixels/dt");
    dataV->add_att("units","pixels/dt");
    //-put variables
    dataX->put(x, nx);
    dataY->put(y, ny);
    dataU->put(tempUV.data(), nx, ny);
    dataV->put(tempUV.data(0,0,1), nx, ny);
    //-free
    delete[] x;
    delete[] y;
  }
  else{
    cout << "Couldn't open file!" << endl;
    err = 2;
  }
  //-return
  return err;
};
#endif

/**
 *  Load .nc files.
 *
 * written on 24/01/2011
 */
#ifdef USE_NCLIB
CImg<double> read_NCDF(const char *const filename){
  
  int nx, ny;
  CImg<float> tempUV;
  
  //-open file
  NcFile dataFile(filename, NcFile::ReadOnly);
  if (dataFile.is_valid())
  {
    //-get dimensions
    nx = dataFile.get_dim("x")->size();
    ny = dataFile.get_dim("y")->size();
    //-allocate image
    tempUV.assign(nx,ny,2);    
    //-load variables (u , v)
    //NcVar *dataU = dataFile.get_var("u");
    //NcVar *dataV = dataFile.get_var("v");
    //-read values
    //dataU->get(UVtemp.data(),nx,ny);
    //dataV->get(UVtemp.data(0,0,1),nx,ny);
    dataFile.get_var("u")->get(tempUV.data(),nx,ny);
    dataFile.get_var("v")->get(tempUV.data(0,0,1),nx,ny);
    //-re orient
    tempUV.get_shared_plane(1) *= -1.;
  }
  else{
    cout << "Couldn't open file, empty field returned ..!" << endl;
  }
  //-return
  return CImg<double>(tempUV,false);
};


CImg<double> read_NCDF_Im(const char *const filename){
    
    int nx, ny, nz;
    CImg<float> temp;
    
    //-open file
    cout<<filename<<endl;

    NcFile dataFile(filename, NcFile::ReadOnly);
    if (dataFile.is_valid())
    {
        //-get dimensions
        nx = dataFile.get_dim("lon")->size();cout<<"nb longitude samples = "<<nx<<endl;
        ny = dataFile.get_dim("lat")->size();cout<<"nb latitude samples = "<<ny<<endl;
        nz = 1;//dataFile.get_dim("time")->size();cout<<"nb time samples = "<<ny<<endl;
        //-allocate image
        temp.assign(nx,ny,nz);
        //-load variable sst
        dataFile.get_var("sea_surface_temperature")->get(temp.data(),nz,ny,nx);
    }
    else{
        cout << "Couldn't open file, empty field returned ..!" << endl;
    }
    //-return
    return CImg<double>(temp,false);
};
#endif


/**
 *  Read .uv
 * re-written on 29/04/2010
 * => simpler code ! 
 * "1 pix delay" bug removed !
 */
CImg<double> read_aUVFile(char * filename,int dx,int dy,int Dx,int Dy, int ofx, int ofy){
  
	int ii,jj;
	double u,v;
  ifstream file(filename, ios::in); 
	CImg<double> im(dx,dy,2,0);
	CImg<double> aux(Dx,Dy, 2);
  
  if (file){
    //--read file
    for (int j=0;j<Dx;j++){
      for (int i=0;i<Dy;i++){
        file >> jj >> ii >> u >> v;
        aux(j,i,0)=u;
        aux(j,i,1)=-v;      
      }
    }
    file.close();
    //--mirror along y
    //--and crop as supposed to be
    im = aux.mirror('y').crop(ofx,ofy,ofx+dx-1,ofy+dy-1);
    cout << "file " << filename << " successfully loaded." << endl;
  }
  else {
    cout << "\t\t/!\\ file not found, 0-filled image returned." << endl;
  }
	return 	im;
};

/**
 *  Write .uv
 * written on 06/05/2010
 * This is a template function (float or double), see .h
 */

/**
 *  Read .motion
 * re-written on 07/05/2010
 *
 * U, V fields are read line by line.
 * 2 versions : CImage or arrays.
 */

CImg<double> read_MotionFile(const char *const filename,
                            int &err){

  int dx=0, dy=0;
  CImg<double> out;
  err = 0;
  //-read dimensions
  read_MotionFileHeader(filename, dx, dy, err);
  if (!err){
    //-create image
    out.assign(dx,dy,2,1,0.);
    //-read data
    read_MotionFileData(filename, dx, dy, out.data(), out.data(0,0,1), err);
    if (err>0){exit(err);}
  }
  return out;
};

void read_MotionFile(const char *const filename,                         
                     int &dx,
                     int &dy,
                     double* &field_u,
                     double* &field_v,
                     int &err){
  err = 0;
  //-read dimensions
  read_MotionFileHeader(filename, dx, dy, err);
  if (err){exit(err);}
  //-allocate arrays
  field_u = new double[dx*dy];
  field_v = new double[dx*dy];
  //-read data
  read_MotionFileData(filename, dx, dy, field_u, field_v, err);
  if (err){exit(err);}
};

void read_MotionFileHeader(const char *const filename,
                           int &dx,
                           int &dy,
                           int &err){
  err = 0;
  //-open file
  ifstream file(filename, ifstream::in);
  if (file){
    //-read header
    file >> dx >> dy;
    file.close();
  }
  else{
    cout << "Impossible to open file [" << filename << "]. Empty field!" << endl;
    err = -1;
  }
};

void read_MotionFileData(const char *const filename,
                         int dx,
                         int dy,
                         double* field_u,
                         double* field_v,
                         int &err){
  
  string fieldnames;
  int ii, jj, nVal;
  
  err = 0;
  //-open file
  ifstream file(filename, ifstream::in);
  if (file){
    cout << "lecture fichier " << filename << ": ";
    //-read header
    file >> ii >> jj;
    cout << "dimensions ( " << ii << "x" << jj << " ) ; ";
    if ((ii!=dx)||(jj!=dy)) {
      cout << "Bad dimensions ...." <<endl;
      exit(1);
    }
    nVal = dx*dy;
    cout << "fields ( ";
    for (int k=0;k<4;k++){
      file >> fieldnames;
      cout << fieldnames << " ";
    }
    cout << ")";
    //-then data
    for (int k=0; k<nVal; k++) file >> ii >> jj >> field_u[k] >> field_v[k];
    cout << " >> Done !" << endl;
  }
  else{
    cout << "Impossible to open file " << filename << ". Exit !" << endl;
    err = -1;
  }
};


/**
 *  Write .motion
 * written on 03/05/2010
 * => simpler code
 *
 * U, V fields are written line by line.
 * This is a template function (float or double), see .h
 */

/**
 *  Read .pgm file
 */
CImg<short> read_pgm(char *nom,
                     int dx,
                     int dy,
                     int ofx,
                     int ofy){
  
  FILE *im;
  im=fopen(nom,"rb");
  char comm;
  char t1[1000];
  int def;
  long int position;
  short *nb_col=new short[1];
  short *nb_lig=new short[1];
  
  /*header*/
  fgets(t1,1000,im);
  fscanf(im,"%c",&comm);
  while(comm == '#'){
    fgets(t1,1000,im);
    fscanf(im,"%c",&comm);
  }
  position=(long int)(ftell(im)-1);
  fseek(im,position,0);
  fscanf(im,"%hd %hd\n%d\n",nb_col,nb_lig,&def);
  
  int Dx=*nb_col;
  CImg<short> image(dx,dy);
  /*body*/
  if(def==255){
    char *imInt;
    imInt=(char*)calloc(*nb_col*(*nb_lig),sizeof(char));
  	fread(imInt,sizeof(char),*nb_col*(*nb_lig),im);
    for (int i=0;i<dx;i++){
      for (int j=0;j<dy;j++){
        image(i,j)=(short)(unsigned char)imInt[i+ofx+(j+ofy)*Dx];
      }
    }
    free(imInt);
  }
  else{
    if (def == 1023){
      fseek(im,position,0);
      fgets(t1,1000,im);
      fgets(t1,1000,im);
      unsigned short *imInt;
      imInt=(unsigned short*)calloc(*nb_col*(*nb_lig),sizeof(unsigned short));
     	fread (imInt, sizeof(unsigned short), *nb_col*(*nb_lig), im );
      for (int i=0;i<dx;i++){
        for (int j=0;j<dy;j++){
          image(i,j)=imInt[i+ofx+(j+ofy)*Dx];
        }
      }
      free(imInt);
    }
    else{
      printf("Can't read images %s UVolution problem def=%d (255 ou 1023)!",nom,def);
      exit(1);
    }
  }
  fclose(im);
  delete nb_col;
  delete nb_lig;
  return image;
};

/**
 *  Read lidar binary file
 * written on 03/05/2010
 */
void read_lidar(char* filename,
                double* output,
                int width,
                int height,
                int &err){
  float buff;
  err = 0.;
  ifstream flidar(filename, ios::in | ios::binary);
  if(flidar.is_open()){
    cout << "lecture fichier " << filename << endl;
    for (int q=0; q<width; q++) {
      for (int p=0; p<height; p++) {
        flidar.read(reinterpret_cast<char*>(&buff),sizeof(float));
        output[p*width+q] = (double)buff;
      }
    }
    flidar.close();
  }
  else {
    cerr << "Impossible d'ouvrir le fichier " << filename << endl;
    err = 1;
  }
  
}



