/*
 *  Anisotropic Fractional Divergence-free aniWavOpticalFlow.h
 *
 *
 *  Written by Patrick Héas.
 *  Copyright 2014 INRIA. All rights reserved.
 *
 */

#ifndef __ANIWAVOPTICALFLOW_H
#define __ANIWAVOPTICALFLOW_H
#include <string>
#include <fstream>
#include "CImg.h"
#include "Daubechies.h"
#include "OpFlowTools.h"
#include "lbfgs.h"
using namespace std;
/*!
  @class WavOpticalFlow
  @abstract This class builds and solves a wavelet-based optic flow problem with a divergence-free fBm prior.
 */

class WavOpticalFlow {
  
  //=================
  //-----MEMBERS-----
  //=================
protected:
  //--------
  //-imagesˇ
  //--------
  /*! @var I0 Image at time t. */
  cimg_library::CImg<double> I0;
  /*! @var I1 Image at time t+1. */  
  cimg_library::CImg<double> I1;
  /*! @var UV Estimated velocity field. */  
  cimg_library::CImg<double> UV;
    /*! @var UV true velocity field. */  
  cimg_library::CImg<double> UVreal;
  /*! @var errEndPoint End-point error map. */  
  cimg_library::CImg<double> errEndPoint;
  /*! @var errBarron Barron error map. */  
  cimg_library::CImg<double> errBarron;
  /*! @var Mask Optional mask to discard irrelevant regions. */  
  cimg_library::CImg<double> Mask;
  /*! @var Weights Robust weights map. */  
  cimg_library::CImg<double> Weights;
  /*! @var gradI UNUSED. */ 
  cimg_library::CImg<double> gradI;
  /*! @var cU U field coefficients pyramid. */
  cimg_library::CImg<double> cU;
  /*! @var cV V field coefficients pyramid. */ 
  cimg_library::CImg<double> cV;
  /*! @var Lambda binary map for inertial scale range. */ 
  cimg_library::CImg<double> Lambda;
  
  /*! @var Normalizing map for foureir-based Div-free wavelets. */ 
  cimg_library::CImg<double> NormMap;
  
  
  cimg_library::CImgList<double> evenMass;
  cimg_library::CImgList<double> oddMass;
  
  cimg_library::CImgList<double> pStiffW;
  cimg_library::CImgList<double> pStiffH;
  cimg_library::CImgList<double> a_pStiffW;
  cimg_library::CImgList<double> a_pStiffH;
 //-------------------
  //-coefficients vector
  //--------------------
  /*! @var theta Vector grouping U and V coefficients. */
  cimg_library::CImg<double> theta;
  cimg_library::CImg<double> thetaReal;
  cimg_library::CImg<double> deltatheta;  
  /*! @var dimW Horizontal problem dimension (x). */
  int dimW;
  /*! @var dimH Vertical problem dimension (y). */
  int dimH;
  //-------
  //-levels
  //-------
  /*! @var n0 Total number of scales (horizontal+vertical). */  
  int n0;
  /*! @var n0w Total number of horizontal scales. */
  int n0w;
  /*! @var n0h Total number of vertical scales. */
  int n0h;
  /*! @var nDecomp Number of considered scales considered (deepness of decomposition). */  
  int nDecomp;
  /*! @var nEstim Number of estimated scales. */
  int nEstim;
  /*! @var nInit Number of initialized scales. */
  int nInit;
  /*! @var currentLvl Current estimated scale. */  
  int currentLvl;
  //------------------------------
  //-total/current number of coeff
  //------------------------------
  /*! @var nTotCoeff Total number of coefficients (per field, so actual total is 2x). */  
  int nTotCoeff;
  /*! @var nCoeff Current number of coefficients, related to currentLvl (per field, so actual total is 2x). */  
  int nCoeff;
  //-------
  //-filter
  //-------
  /*! @var wav Daubechies object for wavelet operations. */
  Daubechies wav;
  //-----------
  //-regularisation
  //-----------
  /*! @var regul Type of regularization. */  
  int regul;
  //-for penalisation
  /*! @var alpha Factor to balance data model and penalization. */  
  double alpha;
  /*! @var gamma Factor to balance data model and fractional prior. */  
  double gamma;
  /*! @var upper scale of power law range. */  
  double MaxPwrRange;
  /*! @var lower scale of power law range. */  
  double MinPwrRange;
  //-----------------
  //-lbfgs parameters
  //-----------------
  /*! @var paramTheta L-BFGS parameters for minimization over theta (coefficients). */  
  lbfgs_parameter_t paramTheta;
  /*! @var paramLambda L-BFGS parameters for minimization over lambda (dual problem). */  
  lbfgs_parameter_t paramLambda;
  //---------
  //-various
  //---------
  /*! @var sigma Value for smoothing. */
  double sigma;
  /*! @var fval Current total cost function value. */
  double fval;  
 /*! @var zeta Fractional exponent. */
  double zeta;
 /*! @var zetaIn Spectrum exponent. */
  double zetaIn;
  /*! @var divFree Boolean, true for divergence free wavelets. */
  bool divFree;
    /*! @var periodic Boolean for border conditions. */
  bool periodic;
  /*! @var fraction Boolean, true for fractional wavelets. */
  bool fraction;
  /*! @var LerayProj Boolean, true for Leray Projection operation. */
    bool LerayProj;
  /*! @var	c standard deviation of iid gaussian fractional wavelet coefficients */
  double c;
  /*! @var	tau coef attenuation for normalization of fractional wavelets*/
  double tau;
    /*! @var approx Boolean, true for approx with reg. approx connection coefficients. */
    bool approx;
/*! @var coeffDir Directory where connection coefficients file are strored. */
    string coeffDir;
  //=================
  //----METHODS----
  //=================
public:
  
  /*!
    @abstract     Problem constructor.
    @param        Im0 image at time t
    @param        Im1 image at time t+1
    @param        ImMask mask for gradients (=>cost function) computation
    @param        nD number of wavelet decompositions (scales) considered on Im0 & Im1
    @param        nE number of estimated scales
    @param        nI number of initialized scales (if initialization field given)
    @param        nM number of wavelet vanishing moments (=> regularity)
    @param        wT Type of wavelet (see Daubechies library)
    @param        sig sigma parameter for smoothing (gradient computation)
    @param        reg regularization type
    @param        a penalization alpha value
   */
  WavOpticalFlow(cimg_library::CImg<double> Im0,
                 cimg_library::CImg<double> Im1,
                 cimg_library::CImg<double> ImMask,
                 int nD,
                 int nE,
                 int nI,                 
                 int nM,
                 int wT,
                 double sig,
                 int reg,
                 double a);
  
    
  /*!
    @abstract     Problem constructor (fractional div-free case).
    @param        Im0 image at time t
    @param        Im1 image at time t+1
   @param        ImMask mask for gradients (=>cost function) computation
   @param      UVtrue true velocity field
    @param        nD number of wavelet decompositions (scales) considered on Im0 & Im1
    @param        nE number of estimated scales
    @param        nI number of initialized scales (if initialization field given)
    @param        nM number of wavelet vanishing moments (=> regularity)
    @param        wT Type of wavelet (see Daubechies library)
    @param        sig sigma parameter for smoothing (gradient computation)
    @param        reg regularization type
    @param        a penalization alpha value
   @param         g penalization gamma value
   @param         z fractional derivative exponent value
   @param         zIn Spectrum exponent value
   @param         frac boolean for fractional wavelet activation
   @param         div0 boolean for div-free wavelet activation
  @param          per boundary condition for interpolation
  @param          minPwr lower scale bound of power law range
  @param          maxPwr upper scale bound of power law range
 @param			 c standard deviation of iid gaussian fractional wavelet coefficients 
 @param			 t coef attenuation for normalization of fractional wavelets 
   */
  WavOpticalFlow(cimg_library::CImg<double> Im0,
                 cimg_library::CImg<double> Im1,
                 cimg_library::CImg<double> ImMask,
                 cimg_library::CImg<double> ImUVreal,
                 int nD,
                 int nE,
                 int nI,                 
                 int nM, 
                 int wT,
                 double sig,
                 int reg,
				 double g,
				 double z,
				 double zIn,
			     bool frac,
			     bool div0,
			     bool per,
				 double minPwr,
				 double maxPwr,
				 double c, 
				 double t,
                 bool app,
                 string coeffDir_);
				 
  /*!
    @abstract Problem destructor (de-allocate).
   */
  ~WavOpticalFlow();
  
  /*!
    @abstract   Display information on current problem.
    @param      disp will display images and mask
   */
  void info(bool disp=false);
  
  
  /*!
    @abstract   Solve a problem with penalization and/or truncated basis, using L2 norm.
    @param      epsilon tolerance vor l-bfgs convergence.
    @param      UV0 initilaization velocity field
    @param      interp interpolated non-estimated scales
    @param      div null-divergence projection
    @result     The estimated UV field.
   */
  cimg_library::CImg<double> solve(double epsilon,
                                   const cimg_library::CImg<double> &UV0,
                                   bool interp=false,
                                   bool div=false);
  
   /*!
    @abstract   Solve a problem (with fractional divergence-free wavelets) with penalization and/or truncated basis, using L2 norm.
    @param      epsilon tolerance vor l-bfgs convergence.
    @param      UV0 initilaization velocity field
	@result     The estimated UV field.
   */
  cimg_library::CImg<double> solveFracDivFree(double epsilon,
                                   const cimg_library::CImg<double> &UV0); 
  
  
  
  /*!
    @abstract   Solve a problem with robust estimator (instead of L2 norm).
    @param      epsilon tolerance vor l-bfgs convergence.
    @param      robustType type of robust estimator
    @param      threshold sigma value of robust estimator
    @param      UV0 initilaization velocity field
    @param      interp interpolated non-estimated scales
    @param      div null-divergence projection
    @result     The estimated UV field.
   */
  cimg_library::CImg<double> solveRobust(double epsilon,
                                         int robustType,
                                         double threshold,
                                         const cimg_library::CImg<double> &UV0,
                                         bool disp=false,
                                         bool interp=false,
                                         bool div=false);
  
  /*!
    @abstract   Estimate a scale.
    @param      deltatheta an array of 2 times of size the current number of U&V coefficents, i.e. 2x @link nCoeff nCoeff @/link
   */
  void stepSolve();
  
  /*!
    @abstract   Estimate a scale in a fractional divergence-free basis.
    @param      deltatheta an array of 2 times of size the current number of U&V coefficents, i.e. 2x @link nCoeff nCoeff @/link
   */
  void stepSolveFracDivFree (); 
  /*!
    @abstract   Initialize the estimated @link UV UV@/link field with given field. Only @link nInit nInit@/link scales initialized.
    @param      UV0 the input velocity field
   */
  void initUV(const cimg_library::CImg<double> &UV0);
 
 /*!
    @abstract   Initialize the estimated @link UV UV@/link field with given field. Only @link nInit nInit@/link scales initialized.
    @param      UV0 the input velocity field
   */
  void initUVFracDivFree(const cimg_library::CImg<double> &UV0);
   
  /*!
    @abstract   Fill velocity field coefficients @link cU cU@/link, @link cV cV@/link with values from given array. Then rebuild @link UV UV@/link field from @link cU cU@/link, @link cV cV@/link. 
    @param      X array containing coefficients values to be used, of size 2xnC
    @param      nC number of coefficients (first index of V coefficients in array)
   */
  void setUV(const double *X,
             int nC);
			 
  	/*!
	 @abstract   Fill velocity field coefficients @link cU cU@/link, @link cV cV@/link with values from given array. Then rebuild @link UV UV@/link field from @link cU cU@/link, @link cV cV@/link. 
	 @param      X array containing coefficients values to be used, of size 2xnC
	 @param      nC number of coefficients (first index of V coefficients in array)
	 */
	void setUVFracDivFree(const double *X,
			   int nC);

 	
    /*!
	 @abstract   Divergence free fractional random field generator. 
	 @result    Random field.
	 */
	cimg_library::CImg<double> generatorUVFracDivFree();
	
 	/*!
	 @abstract   Divergence free fractional random field generator. 
	 @param     zetaGen power law exponent 
	 @result    Random field.
	 */
	cimg_library::CImg<double> generatorUVFracDivFree(double zetaGen);
	cimg_library::CImg<double> generatorUVFracDivFree( double zetaGen, double cc);
	cimg_library::CImg<double> generatorUVFracDivFree( double zetaGen, double cc, double minS, double maxS);
		
 /*!
	 @abstract  Create image transported by divergence free fractional random field generator. 
	 @param     zetaGen power law exponent 
	 @result    Random field and images.
	 */
    cimg_library::CImgList<double>  deformAndReturnImByUVFracDivFree(bool fourier, double zetaGen, double cc, double minS, double maxS,double sigNoise)	;
    cimg_library::CImg<double>  deformImByUVFracDivFree(bool fourier, double zetaGen, double cc, double minS, double maxS);
	cimg_library::CImg<double>  deformImByUVFracDivFree( double zetaGen, double cc);
	cimg_library::CImg<double>  deformImByUVFracDivFree( double zetaGen);
	cimg_library::CImg<double>  deformImByUVFracDivFree();

  /*!
    @abstract   compute (DFD data term) cost function value from current @link UV UV@/link field. 
    @result     Value fo the DFD cost function.
   */
  double eval_Jobs();
  
  /*!
    @abstract   compute (DFD data term) cost function value and gradient from current @link UV UV@/link field. 
    @param      g array of size 2x @link nCoeff nCoeff@/link to be filled with gradient
    @result     Value fo the DFD cost function.
   */  
  double eval_JobsGrad(double *g);
  
  /*!
	 @abstract   compute (DFD data term) cost function value and gradient from current @link UV UV@/link field. 
	 @param      g array of size 2x @link nCoeff nCoeff@/link to be filled with gradient
     @param     x current point where function is evaluated, of size 2x @link nCoeff nCoeff@/link
	 @param      zeta fractional exponent
	 */  
	double eval_JobsGradFracDivFree(double *g,const double *x);

  /*!
    @abstract  Function called by optimisation routine to evaluate cost function and gradient at a given point. Includes observation and eventual penalization term!
    @param     x current point where function is evaluated, of size 2x @link nCoeff nCoeff@/link
    @param     g array to be filled with gradient, of size 2x @link nCoeff nCoeff@/link
    @result    Value of the cost function.
   */
  double lbfgs_eval_CostFunction(const double *x,
                                 double *g);
  	/*!
	 @abstract  Function called by optimisation routine to evaluate cost function and gradient at a given point for divergence-free fractional wavelets. Includes observation and eventual penalization term!
	 @param     x current point where function is evaluated, of size 2x @link nCoeff nCoeff@/link
	 @param     g array to be filled with gradient, of size 2x @link nCoeff nCoeff@/link
	 @result    Value of the cost function.
	 */
	double lbfgs_eval_CostFunctionFracDivFree(const double *x,
											  double *g);
  /*!
    @abstract   Customizable function called by @link _progress _progress@/link l-bfgs routine. Display progress of optimization.
    @param      x the current point
    @param      g gradient at current point
    @param      fx function value at current point
    @param      xnorm
    @param      gnorm
    @param      step step size
    @param      n
    @param      k number of iteration
    @param      ls
    @result     return 0 =).
   */
  int progress(const double *x,
               const double *g,
               const double fx,
               const double xnorm,
               const double gnorm,
               const double step,
               int n,
               int k,
               int ls);
    
  /*!
    @abstract   If ground truth provided, computes several errors.
    @param      trueField ground-truth velocity field
    @param      disp weither to display estimation and truth side by side or not
   */
  void compare(cimg_library::CImg<double>& trueField,
               bool disp=false);
  
  /*!
    @abstract  Write output .motion files.
    @param     outDir path and prefix for output file, e.g. outDir="Case_1/run_3/frame_1" will give Case_1/run_3/frame_1_UV.motion
    @param     Ox horizontal offset to remove artificial borders (wavelets need 2^K lengths)
    @param     Oy vertical offset to remove artificial borders
    @param     dimx actual area horizontal size
    @param     dimy actual area vertical size
   */
  void output(std::string outDir,
              int Ox,
              int Oy,
              int dimx,
              int dimy);

 
  //-----------not commented-----------
  void setMatrices();
   void  setFracMatrices_New();
  void setLambda();
  
  cimg_library::CImg<double> massMat(int order, int nLevel);
  cimg_library::CImg<double> stiffFracMat_New(float minusZeta, int nLevel, int p, bool b);
    
    
  double addHighOrderRegul(const double *x,
                           double *g);
    double addFracGaussPrior_New(const double *x,
                             double *g);
        
 /*!
    @abstract   Compute the fractional regularization term (IID zero mean Gaussian energy) and its gradient. 
    @param      instance a pointer on current WavOpticalFlow instance
    @param      x current point
    @param      g gradient at current point
	@result     Returns value of regularization functional. 
   */
 double addFracDivFreeRegul(int lvl,
						   const double *x,
						   double *g);
						   
 /********************
   *
   *  Static methods
   *
   *******************/

  /*!
    @abstract   Preprocess images to assert correct dimensions (power of 2), normalize data and so on.
    @param      I0in input image 0
    @param      I1in input image 1
    @param      Bgin input background image that is substracted from I0in and I1in
    @param      Maskin input mask
    @param      UV0in input initialization field
    @param      UVTin input ground truth 
    @param      I0out output (preprocessed) image 0
    @param      I1out output (preprocessed) image 1
    @param      Maskout output (preprocessed) mask
    @param      UV0out output (preprocessed) initialization field
    @param      UVTout output (preprocessed) ground truth
    @param      Ox horizontal offset, if dimensions corrected by preprocess
    @param      Oy vertical offset, if dimensions corrected by preprocess   
    @param      transport weither initialization field should be transported at current time
    @param      nuViscos flow viscosity for transport
   */
   
   
   	   						   
  static void prepareInputImg(cimg_library::CImg<double> I0in,
                              cimg_library::CImg<double> I1in,
                              const cimg_library::CImg<double> &Bgin,
                              const cimg_library::CImg<double> &Maskin,
                              const cimg_library::CImg<double> &UV0in,
                              const cimg_library::CImg<double> &UVTin,
                              cimg_library::CImg<double> &I0out,
                              cimg_library::CImg<double> &I1out,
                              cimg_library::CImg<double> &Maskout,
                              cimg_library::CImg<double> &UV0out,
                              cimg_library::CImg<double> &UVTout,                              
                              int &Ox,
                              int &Oy);
  
    
    /*!
    @abstract   Evaluation function for lbfgs routine. 
    @param      instance a pointer on current WavOpticalFlow instance
    @param      x current point
    @param      g gradient at current point
    @param      n
    @result     Return @link lbfgs_eval_CostFunction lbfgs_eval_CostFunction method @/link result. 
   */
  static lbfgsfloatval_t _evaluate(void *instance,
                                   const lbfgsfloatval_t *x,
                                   lbfgsfloatval_t *g,
                                   const int n,
                                   const lbfgsfloatval_t step);
  
  /*!
    @abstract   Display function for lbfgs routine. 
    @param      x current point
    @param      g gradient at current point
    @param      fx function value at current point
    @param      xnorm norm of x
    @param      gnorm norm of g
    @param      step step value
    @param      n
    @param      k
    @param      ls
    @result     Return @link progress progress method @/link result.
   */
  static int _progress(void *instance,
                       const lbfgsfloatval_t *x,
                       const lbfgsfloatval_t *g,
                       const lbfgsfloatval_t fx,
                       const lbfgsfloatval_t xnorm,
                       const lbfgsfloatval_t gnorm,
                       const lbfgsfloatval_t step,
                       int n,
                       int k,
                       int ls);
  
  static inline int period(int ind,
                           int max);
  
};

#endif